/**
  *  \file arch/unix/dirs.cc
  *  \brief Unix Directory Reading
  */

// half-hearted attempt to get this compiled with -ansi
#define _POSIX_C_SOURCE 2
#define _XOPEN_SOURCE 600

#include <cerrno>
#include <cstring>
#include <unistd.h>
#include <sys/stat.h>
#include <dirent.h>
#include "io/dirs.h"
#include "cpluslib/assert.h"
#include "util/except.h"
#include "util/nls.h"

namespace {
    // Unix directory implementation.
    class UnixDirectory : public Directory {
        class Entry;
        class Enum;
        string_t dir;
        friend class Enum;
     public:
        UnixDirectory(string_t aname);
        ~UnixDirectory();
        Ptr<DirectoryEntry> getEntryByName(string_t name);
        Ptr<Enumeration<Ptr<DirectoryEntry> > > getEntries();
        Ptr<Directory> getParent();
        string_t getName();
    };

    class UnixDirectoryWithTitle : public UnixDirectory {
        string_t title;
     public:
        UnixDirectoryWithTitle(string_t name, string_t title)
            : UnixDirectory(name), title(title) { }
        string_t getTitle()
            { return title; }
    };
    
    // Unix root directory enumerator.
    class UnixRootEnum : public Enumeration<Ptr<Directory> > {
        enum { get_Root, get_Home, get_Done };
        int where;
     public:
        UnixRootEnum();
        ~UnixRootEnum();
        bool hasMoreElements();
        Ptr<Directory> getNextElement();
    };
}

/** Get absolute name. Returns the canonical, absolute name for the
    specified directory.
    \post isAbsoluteName(result) */
string_t
Directory::getAbsoluteName(string_t dir)
{
    if (!isAbsoluteName(dir)) {
        string_t result = makeFileName(getWorkingDirectory(), dir);
        ASSERT(isAbsoluteName(result));
        return result;
    } else {
        return getCanonicalName(dir);
    }
}

/** Get current working directory.
    \post isAbsoluteName(result)
    \throws FileIOException on error */
string_t
Directory::getWorkingDirectory()
{
    // FIXME: to handle symlink-awareness of some shells, we should
    // handle the "PWD" environment variable, too. Something like
    //    if ($PWD exists && $PWD and "." refer to same inode)
    //    then return $PWD
    //    else get physical directory name
    // Lusers expect `cd symlink; cat ../file' to work.
    // OTOH, only few programs actually *do* support that consistently
    // (if at all possible):
    //   $ cd cpluslib<RET>
    //   $ head ../acdefs<TAB>
    //   acdefs.h   acdefs.hin   acdefs.mak    acdefs.min
    //   $ head ../acdefs.hin<RET>
    //   head: ../acdefs.hin: No such file or directory
    //   $ cd ..<RET>
    //   $ head acdefs.hin<RET>
    //   .....
    std::size_t len = 0;
    char* buffer;
    
    while (1) {
        len += PATH_MAX;
        buffer = new char[len];
        if (getcwd(buffer, len))
            break;
        if (errno != ERANGE)
            throw FileIOException("<getWorkingDirectory>", std::strerror(errno));
        delete[] buffer;
    }

    string_t result(buffer);
    delete[] buffer;
    return result;
}

/** Create directory object for named directory. */
Ptr<Directory>
Directory::openDir(string_t dir)
{
    return new UnixDirectory(dir);
}

/** Get list of root directories ("drives"). For systems that support
    multiple roots, this returns an enumeration of those. For other
    systems, this may return a list of convenient start paths. */
Ptr<Enumeration<Ptr<Directory> > >
Directory::getRoots()
{
    return new UnixRootEnum();
}

//////////////////////////////////////////////////////////////////////

/**
  *  \class UnixDirectory
  *  \brief Directory on File System
  *
  *  This class represents an actual file system directory.
  */

/** File System Directory Entry. Implementation of DirectoryEntry
    for file system directories. */
class UnixDirectory::Entry : public DirectoryEntry {
    friend class UnixDirectory;
    friend class UnixDirectory::Enum;
    Ptr<UnixDirectory> parent;
    Ptr<Entry> link;
 protected:
    void getInfo();
 public:
    Entry(Ptr<UnixDirectory> fsd, string_t aname);
    Ptr<Directory> openDir();
    Ptr<Stream>    openFile(int mode);
    Ptr<Directory> openParent();
    void erase();
    void renameTo(string_t name);
};

UnixDirectory::Entry::Entry(Ptr<UnixDirectory> fsd, string_t aname)
    : DirectoryEntry(aname), parent(fsd), link(0)
{ }

Ptr<Directory>
UnixDirectory::Entry::openDir()
{
    return new UnixDirectory(name);
}

Ptr<Stream>
UnixDirectory::Entry::openFile(int mode)
{
    return Stream::openFile(name, mode);
}

Ptr<Directory>
UnixDirectory::Entry::openParent()
{
    return parent;
}

void
UnixDirectory::Entry::erase()
{
    if (unlink(name.c_str()) != 0)
        throw FileIOException(name, std::strerror(errno));
    clearInfo();
}

void
UnixDirectory::Entry::renameTo(string_t newname)
{
    string_t newpath = Directory::makeFileName(parent->dir, newname);
    if (rename(name.c_str(), newname.c_str()) != 0)
        throw FileIOException(name, std::strerror(errno));
    clearInfo();
}

void
UnixDirectory::Entry::getInfo()
{
    struct stat st;
    if (lstat(name.c_str(), &st) != 0)
        return;

    FileType ft = S_ISREG(st.st_mode) ? FT_FILE
        : S_ISDIR(st.st_mode) ? FT_DIRECTORY
        : FT_UNKNOWN;
    setInfo(st.st_size, ft);
}

/** File System Directory Enumerator. Implementation of the directory
    enumeration for UnixDirectory. */
class UnixDirectory::Enum : public Enumeration<Ptr<DirectoryEntry> >
{
    Ptr<UnixDirectory> dir;           ///< The directory we're using
    Ptr<UnixDirectory::Entry> entry;  ///< Current entry
    DIR* dirp;                        ///< Current position. 0 when done iterating
    void next();
 public:
    Enum(Ptr<UnixDirectory> adir);
    ~Enum();
    bool hasMoreElements();
    Ptr<DirectoryEntry> getNextElement();
};

UnixDirectory::Enum::Enum(Ptr<UnixDirectory> adir)
    : dir (adir),
      entry (0)
{
    dirp = opendir(dir->dir.c_str());
    if (!dirp)
        throw FileOpenException(dir->dir, std::strerror(errno));
    next();
}

UnixDirectory::Enum::~Enum()
{
    if (dirp)
        closedir(dirp);
}

/** Advance to next entry. Sets entry to the next directory entry. */
void
UnixDirectory::Enum::next()
{
    ASSERT(dirp);

    struct dirent* de;
    do
        de = readdir(dirp);
    while (de && (strcmp(de->d_name, ".") == 0 || strcmp(de->d_name, "..") == 0));
    if (!de) {
        closedir(dirp);
        entry = 0;
        dirp = 0;
    } else {
        entry = new UnixDirectory::Entry(dir, Directory::makeFileName(dir->dir, de->d_name));
    }
}

bool
UnixDirectory::Enum::hasMoreElements()
{
    return entry.ptr();
}

Ptr<DirectoryEntry>
UnixDirectory::Enum::getNextElement()
{
    ASSERT(entry.ptr());
    Ptr<DirectoryEntry> ret = entry;
    next();
    return ret;
}

/** Constructor. Construct a UnixDirectory for the
    directory specified by aname. */
UnixDirectory::UnixDirectory(string_t aname)
    : dir(getCanonicalName(aname))
{ }

UnixDirectory::~UnixDirectory()
{ }

Ptr<DirectoryEntry>
UnixDirectory::getEntryByName(string_t name)
{
    return new Entry(this, makeFileName(dir, name));
}

Ptr<Enumeration<Ptr<DirectoryEntry> > >
UnixDirectory::getEntries()
{
    return new Enum(this);
}

Ptr<Directory>
UnixDirectory::getParent()
{
    string_t my_abs = getAbsoluteName(dir);
    if (my_abs.length() <= 1)
        return 0;
    else
        return new UnixDirectory(makeFileName(my_abs, ".."));
}

string_t
UnixDirectory::getName()
{
    return dir;
}

/****************************** UnixRootEnum *****************************/

/** \class UnixRootEnum
    Enumerates Unix root directories.

    \todo right now, this returns only "/". It should return a list of
    mount points, plus the user's home */
UnixRootEnum::UnixRootEnum()
    : where(0)
{ }

UnixRootEnum::~UnixRootEnum()
{ }

bool
UnixRootEnum::hasMoreElements()
{
    return where < get_Done;
}

Ptr<Directory>
UnixRootEnum::getNextElement()
{
    switch (where++) {
     case get_Root:
        return new UnixDirectoryWithTitle("/", _("Root directory"));
     case get_Home: {
        char* name = std::getenv("HOME");
        if (!name)
            name = "/";
        return new UnixDirectoryWithTitle(name, _("Home directory"));
     }
     default:
        return 0;
    }
}
