/**
  *  \file client/act-planet.cc
  *  \brief Planet Actions
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/dlg-cargo.h"
#include "client/dlg-fcode.h"
#include "client/dlg-list.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "ui/commands.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "game/player.h"
#include "game/action.h"
#include "gfx/palette.h"
#include "client/wcost.h"
#include "client/dlg-tech.h"
#include "game/tables.h"

/** Friendly code change, UI action. */
void
doPlanetFCodeChange(GUniverse& univ, GPlanet& planet)
{
    GFCodeList list(GFCodeList::getInstance(), planet);
    WFCodeWindow dialog(_("Change Friendly Code"), list);
    dialog.setFCode(planet.getFCode());
    ui_root->centerWidget(dialog);
    ui_root->add(dialog);
    if (dialog.run() == cm_Enter) {
        planet.setFCode(dialog.getFCode());
        univ.doScreenUpdates();
    }
}

/** Cargo transfer to/from planet, UI action.
    \param planet    invoking planet
    \param unload    true if 'u' key, false if 'c' key. These have no
                     difference other than the heading of the target
                     selection window. */
void
doPlanetTransfer(GUniverse& univ, GPlanet& planet, bool unload)
{
    GObjectList list;
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& ship = univ.getShip(i);
        if (ship.isExistant() && ship.getOwner() == planet.getOwner() && ship.getPos() == planet.getPos())
            list.addObject(univ.getShipType(), ship);
    }
    if (list.getSize() == 0) {
        messageBox(_("There's none of our ships orbiting this planet."),
                   _("Cargo Transfer"));
        return;
    }
    WObjectSelectionList wlist(0, 15, 320, list);
    GObjectId chosen = wlist.doStandardDialog(unload
                                              ? _("Unload ship...")
                                              : _("Transfer cargo to..."),
                                              0 /* FIXME */);
    if (chosen.isValid())
        doPlanetTransferFor(univ, planet, chosen);
}

/** Planet cargo transfer to ship, UI action.
    \param planet  invoking planet
    \param chosen  target. Must be a ship. */
void
doPlanetTransferFor(GUniverse& univ, GPlanet& planet, GObjectId target)
{
    GMultiTransaction mt;
    GPlanetTransfer& left = mt.addNewSubtransaction(new GPlanetTransfer(planet));
    GShipTransfer&  right = mt.addNewSubtransaction(new GShipTransfer(*checked_cast<GShip*>(&target.getObject())));
    GCargoTransfer&    ct = mt.addNewSubtransaction(new GCargoTransfer(left, right));

    doCargoTransfer(mt, ct);

    univ.doScreenUpdates();
}

/** Starbase mission change, UI action. */
void
doBaseChangeMission(GUniverse& univ, GPlanet& planet)
{
    UIStringList data;
    for (int i = 0; i <= 6; ++i)
        data.add(i, getBaseMissionName(i));

    UIStandardListbox lb(0, 7, 300, data);
    lb.scrollToKey(planet.getBaseMission());
    if (lb.doStandardListbox(_("Starbase Mission"), 0 /* FIXME: help id */)) {
        planet.setBaseMission(lb.getKey());
        univ.doScreenUpdates();
    }
}

/** Starbase shipyard order, UI action.
    \param planet the starbase we're talking about.
    \param w      triggering widget, used for binding the menu. */
void
doBaseFixRecycle(GUniverse& univ, GPlanet& planet, UIWidget& w)
{
    GStarbaseFixRecycleAction action(univ, planet);
    UIStringList d;
    if (action.canFix())
        d.add(sya_Fix, _("Fix (repair) a ship"));
    if (action.canRecycle())
        d.add(sya_Recycle, _("Recycle a ship"));
    if (planet.getBaseShipyardOrder() == sya_Recycle)
        d.add(sya_None, _("Cancel `recycle' order"));
    if (planet.getBaseShipyardOrder() == sya_Fix)
        d.add(sya_None, _("Cancel `fix' order"));

    if (d.getCount() == 0) {
        messageBox(_("There is no ship at this starbase which could be fixed or recycled."),
                   _("Shipyard"));
        return;
    }

    UIStandardListbox lb(0, 1, d.getWidth(), d);
    if (!lb.doStandardMenu(w))
        return;
    switch (int cmd = lb.getKey()) {
     case sya_Recycle:
     case sya_Fix:
        {
            GObjectList list;
            action.enumerateShipsFor(TShipyardAction(cmd), list);
            list.postprocessUser(true);

            const char* title = (cmd == sya_Recycle ? _("Recycle ship") : _("Fix ship"));
            GObjectId chosen = WObjectSelectionList(0, -1, 300, list).doStandardDialog(title, 0 /* FIXME */);
            if (chosen.isValid())
                action.setAction(TShipyardAction(cmd), chosen.getId());
        }
        break;
     case sya_None:
        action.setAction(sya_None, 0);
        break;
    }
    univ.doScreenUpdates();
}

void
doPlanetBuildBase(GUniverse& univ, GPlanet& planet)
{
    GMultiTransaction mt;
    GCargoContainer& pay =
        mt.addNewSubtransaction(new GPlanetTransfer(planet));
    GPlanetBuildStarbaseAction& action =
        mt.addNewSubtransaction(new GPlanetBuildStarbaseAction(planet, pay, !planet.isBuildingBase()));

    if (GError e = action.getError()) {
        if (!e.isError(e.eNoResource)) {
            messageBox(e.getUserError(), _("Build Starbase"));
            mt.doCancel();
            return;
        }
    }

    if (planet.isBuildingBase()) {
        bool answer = UIMessageBox(_("You wanted to build a starbase at this planet. "
                                     "Cancel this order?"),
                                   _("Build Starbase")).doYesNoDialog();
        if (answer && !action.getError())
            mt.doCommit();
        else
            mt.doCancel();
        univ.doScreenUpdates();
    } else {
        static const UIKeyTranslator::Init xl[] = {
            { SDLK_RETURN, cm_Enter },
            { SDLK_ESCAPE, cm_Escape }
        };

        UIWindow win(_("Build Starbase"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
        win.setLayout(UIVBoxLayout::instance5);
        ObjectHolder h;
        win.add(h.add(new WCostDisplay(action,
                                       GCargoTypeSet() + el_Tritanium + el_Duranium + el_Molybdenum + el_Money)));
        UIGroup bg(UIHBoxLayout::instance5);
        bg.add(h.add(new UISpacer()));
        if (!action.getError()) {
            UIKeyString y(_("Yes")), n(_("No"));
            win.add(h.add(new UIStaticText(_("Build this starbase?"), UIStaticText::BlackTitle)));
            bg.add(h.add(new UIButton(cm_Enter,  y.getString(), y.getKey())).addStop());
            bg.add(h.add(new UIButton(cm_Escape, n.getString(), n.getKey())).addStop());
        } else {
            win.add(h.add(new UIStaticText(_("You can't build a starbase here."), UIStaticText::BlackTitle)));
            bg.add(h.add(new UIButton(cm_Escape, _("OK"), SDLK_RETURN)).addStop());
        }
        bg.add(h.add(new UISpacer()));
        win.add(bg);
        win.add(h.add(new UIKeyTranslator(xl, countof(xl))));
        win.pack();
        win.setState(win.st_Modal, true);

        ui_root->centerWidget(win);
        ui_root->add(win);
        int rv = ui_root->run();
        ui_root->close(win);
        if (rv == cm_Enter && !action.getError())
            mt.doCommit();
        else
            mt.doCancel();
        univ.doScreenUpdates();
    }
}
