/**
  *  \file client/dlg-mbox.cc
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-mbox.h"
#include "ui/util.h"
#include "cpluslib/format.h"
#include "ui/window.h"
#include "util/nls.h"
#include "ui/commands.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "cpluslib/string.h"
#include "ui/msgbox.h"
#include "client/dlg-receiver.h"
#include "client/screens.h"
#include "client/dlg-subject.h"

// FIXME: there should be a way to clear the when user leaves his race
static string_t current_search_str;

/**************************** WMessageDisplay ****************************/

/** \class WMessageDisplay
    \brief Widget displaying contents of a mailbox

    This widget displays one message of a mailbox. It automatically
    updates when the user chooses another message. */

WMessageDisplay::WMessageDisplay(int id, GMessagebox& mbox)
    : UIWidget(id), mbox(mbox)
{
    conn = mbox.sig_sel_changed.add(static_cast<UIWidget*>(this),
                                    (void (UIWidget::*)()) &UIWidget::drawWidget);
}

WMessageDisplay::~WMessageDisplay()
{ }

// FIXME: better have this elsewhere?
bool
WMessageDisplay::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key && e.key.code == SDLK_TAB) {
        GMessageFilter FIXME;
        WSubjectList subjlist(1000, mbox, FIXME);
        subjlist.selectMessage(mbox.getCurrent());
        if (subjlist.doStandardDialog())
            mbox.setCurrent(subjlist.getSelectedMessage());
        return true;
    } else {
        return UIWidget::handleEvent(e, second_pass);
    }
}

void
WMessageDisplay::getLayoutInfo(LayoutInfo& info)
{
    info.flags = li_Flexible;
    info.min_size = info.pref_size = font_sizes[FONT_FIXED].scaledBy(41, 21);
}

void
WMessageDisplay::draw(GfxCanvas& can)
{
    UISkin& s = getSkin();
    s.startDraw();
    s.getParameters().font = fonts[FONT_FIXED];
    s.getParameters().bgcolor = GFX_TRANSPARENT;
    s.setColor(UIColor::tc_White);  // FIXME?
    s.drawBackground(can, getExtent());
    outTextFormattedF(can, s, getExtent().x, getExtent().y,
                      getExtent().w, getExtent().h / font_heights[FONT_FIXED],
                      mbox.getText(mbox.getCurrent()));

    string_t indicator = format(" %d/%d", mbox.getCurrent()+1, mbox.getCount());
    int indi_width = fonts[FONT_NORMAL]->getTextWidth(indicator);

    s.outText(can, FONT_NORMAL,
              getExtent().x + getExtent().w - indi_width,
              getExtent().y,
              indicator);

    can.update(getExtent());
}

void
WMessageDisplay::next(int)
{
    mbox.setCurrent(mbox.getCurrent()+1);
}

void
WMessageDisplay::previous(int)
{
    mbox.setCurrent(mbox.getCurrent()-1);
}

void
WMessageDisplay::searchFromMessage(int nr)
{
    string_t substr = strUCase(current_search_str);
    while (nr < mbox.getCount()) {
        if (strUCase(mbox.getText(nr)).find(substr) != string_t::npos) {
            mbox.setCurrent(nr);
            return;
        }
        ++nr;
    }
    messageBox(_("Search text not found."), _("Search in messages"));
}

void
WMessageDisplay::search(int)
{
    if (UIInputLine::run(current_search_str, 40, _("Search in messages"),
                         _("Search for:"), 20, 0)
        && current_search_str.length())
    {
        searchFromMessage(0);
    }
}

void
WMessageDisplay::searchNext(int)
{
    if (current_search_str.length())
        searchFromMessage(mbox.getCurrent()+1);
    else
        search(0);
}

void
WMessageDisplay::forwardMessage()
{
    WReceiverSelector sel(0);
    if (sel.doSimpleDialog(_("Forward Message"))) {
        // FIXME: doSendMessage returns a screen to go to.
        doSendMessage("--- Forwarded Message ---\n" + mbox.getText(mbox.getCurrent()) + "\n--- End Forwarded Message ---",
                      sel.getReceivers());
    }        
}

namespace {
    class WMessageDialog : public UIWindow {
        WMessageDisplay& dpy;
     public:
        WMessageDialog(WMessageDisplay& dpy)
            : UIWindow(_("Messages"), UIWindow::BLUE_WINDOW, COLOR_BLACK),
              dpy(dpy)
            { }

        bool handleCommand(const UICommand& cmd, UIWidget& sender);
    };
}

bool
WMessageDialog::handleCommand(const UICommand& cmd, UIWidget& sender)
{
    if (cmd.code == cm_Prev) {
        dpy.previous(0);
        return true;
    }
    if (cmd.code == cm_Next) {
        dpy.next(0);
        return true;
    }
    return UIWindow::handleCommand(cmd, sender);
}


void
WMessageDisplay::doStandardDialog()
{
    static const UIKeyTranslator::Init key_data[] = {
        SDLK_UP,       cm_Prev,
        SDLK_PAGEUP,   cm_Prev,
        SDLK_DOWN,     cm_Next,
        SDLK_PAGEDOWN, cm_Next
    };

    WMessageDialog win(*this);
    win.setLayout(UIVBoxLayout::instance5);
    win.add(*this);

    UIGroup g(UIHBoxLayout::instance5);
    win.add(g);

    UIButton prev(cm_Prev,  "\036", '-');
    UIButton next(cm_Next,  "\037", '+');
    UIButton search(0, "S", 's');
    UIButton snext(0, "N", 'n');
    UIButton forwd(0, "F", 'f');
    UIButton ok(cm_Enter, _("Close"), SDLK_ESCAPE);
    UIKeyTranslator keys(key_data, countof(key_data));
    UISpacer sp;

    ok.addStop();

    g.add(prev);
    g.add(next);
    g.add(search);
    g.add(snext);
    g.add(forwd);
    g.add(sp);
    g.add(ok);
    g.add(keys);

    prev.addCommand();
    next.addCommand();
    search.sig_fired.add(this, &WMessageDisplay::search);
    snext.sig_fired.add(this, &WMessageDisplay::searchNext);
    forwd.sig_fired.add(this, &WMessageDisplay::forwardMessage);

    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    ui_root->run();
    ui_root->close(win);
}
