/**
  *  \file game/check-fizz.cc
  *  \brief FIZZ.BIN Checksum Access
  *
  *  This module provides functions to access the checksum part of
  *  FIZZ.BIN. This information mirrors the information stored in the
  *  GEN file (see game/gen.cc) and serves as an additional
  *  consistency check for Tim's software. We only write it, but don't
  *  validate it. The registration part is handled in game/reg.cc, not
  *  here.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/checksum.h"
#include "game/gamedirs.h"
#include "util/nls.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "util/except.h"
#include "io/io.h"

/** Checksum part of FIZZ.BIN. */
static int32 fizz[11*3];

/** Adjust checksums. The checksums stored in FIZZ.BIN are adjusted by
    this amount relative to GENx.DAT checksums. Our interface handles
    the adjustment internally. */
static const int fizz_adjust[3] = { 667, 1667, 1262 };

/** Get FIZZ.BIN Checksum File Entry.
    \param otyp   Object type to query
    \param player Player number to query
    \return Adjusted file entry. This is the same as GGen(...).getChecksum(player). */
int32
getFizzEntry(GControlSection otyp, int player)
{
    ASSERT(otyp >= 1 && otyp <= 3);
    ASSERT(player > 0 && player <= 11);

    return fizz[3*(player-1) + (otyp-1)] - fizz_adjust[otyp-1];
}

/** Set FIZZ.BIN Checksum File Entry.
    \param otyp   Object type to query
    \param player Player number to query
    \param checksum Adjusted file entry. Should be the same as GGen(...).getChecksum(player). */
void
setFizzEntry(GControlSection otyp, int player, int32 checksum)
{
    ASSERT(otyp >= 1 && otyp <= 3);
    ASSERT(player > 0 && player <= 11);

    fizz[3*(player-1) + (otyp-1)] = checksum + fizz_adjust[otyp-1];
}

/** Load FIZZ.BIN file. Prepares data structure to use with
    getFizzEntry()/setFizzEntry().
    \throw FileProblemException on error */
void
loadFizz()
{
    Ptr<Stream> f = game_file_dir->openFile("fizz.bin", Stream::C_READ);

    char buffer[sizeof(fizz)];
    if (f->read(buffer, sizeof(buffer)) != sizeof(buffer))
        throw FileTooShortException(*f);

    for (unsigned int i = 0; i < countof(fizz); ++i)
        fizz[i] = getInt32(buffer + sizeof(int32)*i);
}

/** Save FIZZ.BIN file. Stores changes done with setFizzEntry() back
    to disk.
    \throw FileProblemException on error */
void
saveFizz()
{
    Ptr<Stream> f = game_file_dir->openFile("fizz.bin", Stream::C_WRITE);

    char buffer[sizeof(fizz)];
    for (unsigned int i = 0; i < countof(fizz); ++i)
        storeInt32(buffer + sizeof(int32)*i, fizz[i]);

    if (f->write(buffer, sizeof(buffer)) != sizeof(buffer))
        throw FileIOException(*f, _("Write error"));
}
