/**
  *  \file game/player.cc
  *  \brief Player Properties and Accessors
  *
  *  This module contains routines to access and manage properties of
  *  players, including race names.
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <string>
#include "game/player.h"
#include "game/global.h"
#include "util/except.h"
#include "io/io.h"
#include "gfx/palette.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/recode.h"
#include "cpluslib/misc.h"
#include "game/reg.h"

/*! \class GPlayerBitMatrix
    \brief Bit Matrix of Players

    This class provides a matrix of bits, such as is used for alliance
    relations.

    For ease of reference, parameters are called "subject" and
    "object" here. For example, "allies.get(2,9)" asks whether player
    2 has offered an alliance to player 9. */

/** Constructor. */
GPlayerBitMatrix::GPlayerBitMatrix()
{ }

/** Get one bit.
    \param subj Subject
    \param obj  Object
    \return bit value; false if parameters are out of range. */
bool
GPlayerBitMatrix::get(int subj, int obj) const
{
    if (subj <= 0 || subj > countof(data))
        return false;
    if (obj <= 0 || obj > countof(data))
        return false;

    return data[subj-1].contains(obj);
}

/** Set one bit.
    \param subj  Subject
    \param obj   Object
    \param value New value */
void
GPlayerBitMatrix::set(int subj, int obj, bool value)
{
    if (subj <= 0 || subj > countof(data))
        return;
    if (obj <= 0 || obj > countof(data))
        return;

    if (value)
        data[subj-1] += obj;
    else
        data[subj-1] -= obj;
}

/** Get one row.
    \param subj Subject
    \return set of all objects for this subject. */
GPlayerSet
GPlayerBitMatrix::getRow(int subj) const
{
    if (subj <= 0 || subj > countof(data))
        return GPlayerSet();
    return data[subj-1];
}

/** Clear this set. */
void
GPlayerBitMatrix::clear()
{
    for (int i = 0; i < countof(data); ++i)
        data[i].clear();
}

/** List of player characters. The abbreviation for player N in a
    friendly code is player_chars[N]. */
const char player_chars[] = "0123456789ABX";

/** Definition of the race.nm file. */
enum {
    RN_LONG_ORIG  = 0,
    RN_SHORT_ORIG = RN_LONG_ORIG + 30*11,
    RN_ADJ_ORIG   = RN_SHORT_ORIG + 20*11,
    RN_FILE_SIZE  = RN_ADJ_ORIG + 12*11
};

/** Global race name data. */
GRaceNameList racename_data;

/***************************** GRaceNameList *****************************/

/** Default constructor. */
GRaceNameList::GRaceNameList()
{ }

/** Initialize race name list.
    \param s stream pointing to a 'race.nm' file (race names will be
             loaded from that stream) */
GRaceNameList::GRaceNameList(Stream& s)
{
    load(s);
}

GRaceNameList::~GRaceNameList()
{ }

/** Load race names from specified 'race.nm' file.
    \throws FileFormatException if file too short. */
void
GRaceNameList::load(Stream& s)
{
    char buffer[RN_FILE_SIZE];
    if (s.read(buffer, sizeof(buffer)) != sizeof(buffer))
        throw FileTooShortException(s);
    Recoder::getE2I().recodeArray(buffer, sizeof buffer);

    for(int i = 1; i <= 11; ++i) {
        long_names[i]  = getBasicStringN(&buffer[RN_LONG_ORIG  + 30*(i-1)], 30);
        short_names[i] = getBasicStringN(&buffer[RN_SHORT_ORIG + 20*(i-1)], 20);
        adjectives[i]  = getBasicStringN(&buffer[RN_ADJ_ORIG   + 12*(i-1)], 12);
    }

    long_names[0]  = _("Nobody");
    short_names[0] = _("Nobody");
    adjectives[0]  = _("unowned");

    long_names[12]  = _("The Alien Marauder Alliance");
    short_names[12] = _("Alien Marauders");
    adjectives[12]  = _("Alien");
}

/** Get short race name. Returns the short (20 character) name of
    the specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getShortName(int player)
{
    ASSERT(player >= 0 && player <= 12);
    return short_names[player];
}

/** Get full race name. Returns the full (30 character) name of
    the specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getFullName(int player)
{
    ASSERT(player >= 0 && player <= 12);
    return long_names[player];
}

/** Get race adjective. Returns the adjective (12 characters) of the
    specified player. Also handles unowned (0) and alien (12). */
const string_t&
GRaceNameList::getAdjName(int player)
{
    ASSERT(player >= 0 && player <= 12);
    return adjectives[player];
}

/** Expand race names.
    "%N" expands to race N's short name.
    "%-N" expands to race N's adjective.
    "%%" is a percent sign. */
string_t
GRaceNameList::expandNames(string_t text)
{
    string_t rv;

    while (text.size()) {
        string_t::size_type n = text.find('%');
        if (n == text.npos || n+1 >= text.length()) {
            rv += text;
            break;
        }
        rv.append(text, 0, n);
        text.erase(0, n+1);

        bool adj = false;
        if (text.size() > 0 && text[0] == '-') {
            adj = true;
            text.erase(0, 1);
        }
        if (text.size() > 0) {
            if (text[0] == '%') {
                rv += '%';
            } else {
                int race = getPlayerIdFromChar(text[0]);
                if (race > 0 && race <= NUM_OWNERS) {
                    if (adj)
                        rv += getAdjName(race);
                    else
                        rv += getShortName(race);
                }
            }
            text.erase(0, 1);
        }
    }
    return rv;
}

/**************************** Player Id Status ***************************/

/** Current player id. The client presents the game status from this
    player's point of view. \see getPlayerId() */
static int player_id;

/** Current real player id. This is the number of the player we're
    authenticated as. \see getRealPlayerId() */
static int real_player_id;

/** Get current player. That's the one whose data we're looking at,
    0 if none. */
int
getPlayerId()
{
    return player_id;
}

/** Get current authenticated player. That's the one whose password
    we have. May be different from current player if he used a password
    cookie to become somebody else. */
int
getRealPlayerId()
{
    return real_player_id;
}

/** Parse player character into player number. This is used for
    subspace messages, fcodes.cc, etc.
    \returns player id [0,11], or -1. */
int
getPlayerIdFromChar(char c)
{
    if (c >= '0' && c <= '9')
        return c - '0';
    else if (c == 'A' || c == 'a')
        return 10;
    else if (c == 'B' || c == 'b')
        return 11;
    else
        return -1;
}

/** Parse list of players from list of letters. Parses a "123456789ab"-style
    string up to its end or a ',', and leaves the parse position there.
    Used for parsing mission definitions etc.
    \param text [in] Text to parse
    \param index [in/out] Parse position
    \return player set */
GPlayerSet
parsePlayerList(const string_t& text, string_t::size_type& index)
{
    // FIXME: can we merge this with misc.cc:parseFlags?
    GPlayerSet result;
    while (index < text.size() && text[index] != ',') {
        int p = getPlayerIdFromChar(text[index]);
        if (p > 0)
            result += p;
        ++index;
    }
    return result;
}

/** Parse list of player numbers. Parses a "1 2 3 4 5"-style list.
    Used to parse command messages.
    \param text [in] Text to parse
    \return player set */
GPlayerSet
parsePlayerSetFromNumberList(const string_t& text)
{
    GPlayerSet result;
    int tmp = 0;
    for (string_t::size_type index = 0; index < text.size(); ++index) {
        if (text[index] >= '0' && text[index] <= '9') {
            if (tmp < 99)
                tmp = 10*tmp + text[index] - '0';
        } else {
            if (tmp > 0 && tmp <= NUM_OWNERS)
                result += tmp;
            tmp = 0;
        }
    }
    if (tmp > 0 && tmp <= NUM_OWNERS)
        result += tmp;
    return result;
}

/** Format player set into list of numbers.
    \param set player set to format
    \return string
    \inv parsePlayerSetFromNumberList(formatPlayerSetToNumberList(x)) == x */
string_t
formatPlayerSetToNumberList(GPlayerSet set)
{
    string_t result;
    for (int i = 1; i <= NUM_OWNERS; ++i)
        if (set.contains(i))
            result += format(" %d", i);
    if (result.size())
        result.erase(0, 1);
    return result;
}

/** Set player Ids.
    \param eff   "effective" id. The one in whose name we're giving commands.
    \param real  "real" id. The one which was authenticated. */
void
setPlayerIds(int eff, int real)
{
    real_player_id = real;
    player_id = eff;
}

/** Get relation to player n. */
GPlayerRelation
getPlayerRelation(int n)
{
    if (n == player_id)
        return is_Me;
    else
        // FIXME: allies
        return is_Enemy;
}

/** Check whether to allow registered actions for a player.
    So far we only have one registration, but let's keep the door open for
    when we load multiple RSTs of different state (yeah, that's funky!). */
bool
allowRegistered(int player)
{
    (void) player;
    return getRegInfo().isRegistered();
}

/** Load Race Names. Loads the RACE.NM file. */
void
initRaceNames()
{
    Ptr<Stream> file = spec_file_dir->openFile("race.nm", Stream::C_READ);
    racename_data.load(*file);
}
