/**
  *  \file game/ship.cc
  *  \brief Ships
  *
  *  (c) 2001-2008 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/ship.h"
#include "util/console.h"
#include "game/spec.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "game/mission.h"
#include "game/config.h"
#include "game/formula.h"

/******************************* Utilities *******************************/

static const TShipXY shipxy_init = { 0, 0, 0, 0 };

static const TShip blank_ship = {
    0, 0,                                       // ID, Owner
    { UNKNOWNC, UNKNOWNC, UNKNOWNC },           // FCode
    UNKNOWNP,                                   // Warp
    0, 0,                                       // Waypoint-DX, DY
    UNKNOWNP, UNKNOWNP,                         // X, Y
    UNKNOWNP,                                   // Engine
    UNKNOWNP,                                   // Hull
    UNKNOWNP, UNKNOWNP,                         // Beam Type, Count
    UNKNOWNP,                                   // Bay Count
    UNKNOWNP,                                   // Torp Type
    UNKNOWNP,                                   // Ammo
    UNKNOWNP,                                   // Torp Launcher Count
    UNKNOWNP,                                   // Mission
    UNKNOWNP,                                   // PE
    UNKNOWNP,                                   // Tow
    UNKNOWNP, UNKNOWNP,                         // Damage, Crew
    UNKNOWNP,                                   // Colonists
    { 0 },                                      // Name
    { UNKNOWNP, UNKNOWNP, UNKNOWNP, UNKNOWNP }, // Ore
    UNKNOWNP,                                   // Supplies
    { { 0, 0, 0, 0 }, 0, 0, 0 },                // Unload
    { { 0, 0, 0, 0 }, 0, 0, 0 },                // Transfer
    UNKNOWNP,                                   // Intercept
    UNKNOWNP,                                   // Cash
};

static void
copyTargetToShip(TShip& ship, const TShipTarget& target)
{
    ship.owner       = target.owner;
    ship.warp        = target.warp;
    ship.x           = target.x;
    ship.y           = target.y;
    ship.waypoint_dx = 0;
    ship.waypoint_dy = 0;
    ship.hull        = target.hull;
    std::memcpy(ship.name, target.name, sizeof(target.name));
}

GShip::MapInfo::MapInfo(int id)
    : id(id), pos(), owner(0), scanned_mass(0), orbit(0), rc_flag(0), kind(NoShip)
{ }

/********************************* GShip *********************************/

/** Constructor. Creates a blank ship object for ship /id/. */
GShip::GShip(int id)
    : map_info(id), ship_data(0), target_data(0),
      src_flags(), target_flags()
{
    transporter_in_use[0] = transporter_in_use[1] = false;
}

GShip::~GShip()
{
    delete ship_data;
    delete target_data;
}

TShip&
GShip::createShip()
{
    if (!ship_data) {
        ship_data = new TShip(blank_ship);
        ship_data->ship_id = map_info.id;
        // FIXME: give it a name?
        if (map_info.kind == NoShip)
            map_info.kind = HistoryShip;
    }
    return *ship_data;
}

const TShip&
GShip::getDisplayedShip() const
{
    return ship_data ? *ship_data : blank_ship;
}

/* Initialisation */

/** Add SHIP.DAT entry. This assumes that if we see a ship through
    several .dat files, we get the same one each time.
    \param dat  [in] Parsed .dat file entry
    \param srcf [in] Source flags, i.e. bitfield of which file this record is from */
void
GShip::addShipData(const TShip& dat, GPlayerSet srcf)
{
    if (!ship_data)
        ship_data = new TShip(dat);
    src_flags |= srcf;
}

/** Add TARGET.DAT entry. This assumes that if we see a ship through
    several .dat files, we get the same one each time. This does, however,
    handle the case that one record is seen through PHost's ship name filter
    and the other one isn't, and uses the better file name.
    \param target [in] Parsed TARGET.DAT file entry
    \param srcf   [in] Source flags, i.e. bitfield of which file this record is from */
void
GShip::addTargetData(const TShipTarget& target, GPlayerSet srcf)
{
    if (!target_data) {
        /* creating a new target */
        target_data = new TShipTarget(target);
    } else {
        /* if the new target is better than the one we have, use that */
        if (isDummyName(target_data->name, map_info.id) && !isDummyName(target.name, map_info.id))
            *target_data = target;
    }
    target_flags |= srcf;
}

/** Add SHIPXY.DAT entry. */
void
GShip::addShipXYData(TShipXY xy)
{
    map_info.pos.x = xy.x;
    map_info.pos.y = xy.y;
    map_info.owner = xy.owner;
    map_info.scanned_mass = xy.mass;
}

/** Add remote control flag.
    \param flag=0 for non-RC'd ships
    \param flag<0 if RC is forbidden
    \param flag>0 if to report the ship's real owner */
void
GShip::addRCEntry(int flag)
{
    map_info.rc_flag = flag;
}

/** Get ship data. */
void
GShip::getShipData(TShip& the_data) const
{
    ASSERT(ship_data);
    the_data = *ship_data;
}

/* GMapObject interface */

string_t
GShip::getName() const
{
    // FIXME: handle unknown/blank name
    if (ship_data)
        return getBasicString(ship_data->name);
    else
        return "<unknown>";
}

string_t
GShip::getFullName() const
{
    if (ship_data)
        return format(_("%s (Ship #%d)"), getName(), map_info.id);
    else
        return format(_("Ship %d"), map_info.id);
}

int
GShip::getId() const
{
    return map_info.id;
}

int
GShip::getOwner() const
{
    return map_info.owner;
}

/** Check ship existance. Returns true iff the ship is visible on the
    map (i.e. exists for us at this point in time) */
bool
GShip::isExistant() const
{
    return map_info.owner != 0;
}

GPoint
GShip::getPos() const
{
    return map_info.pos;
}

/* Checking */

void
GShip::internalCheck()
{
    /* figure out what kind we are */
    if (src_flags.nonempty()) {
        /* current ship */
        ASSERT(ship_data != 0);
        map_info.kind = CurrentShip;

        /* check that our other data agrees */
        if (map_info.pos.x != ship_data->x || map_info.pos.y != ship_data->y || map_info.owner != ship_data->owner) {
            /* FIXME: should not warn when we're reading host data */
            console.write(LOG_WARN, format(_("Ship #%d has inconsistent XY information - correcting."), map_info.id));
            map_info.pos.x = ship_data->x;
            map_info.pos.y = ship_data->y;
            map_info.owner = ship_data->owner;
        }
        if (target_data != 0) {
            if (target_data->x != ship_data->x || target_data->y != ship_data->y || target_data->owner != ship_data->owner) {
                console.write(LOG_WARN, format(_("Ship #%d has inconsistent target information - deleting."), map_info.id));
                delete target_data;
                target_data = 0;
            }
        }
    } else if (target_flags.nonempty()) {
        /* current target */
        ASSERT(target_data != 0);
        map_info.kind = CurrentTarget;

        /* check that our other data agrees */
        if (map_info.pos.x != target_data->x || map_info.pos.y != target_data->y || map_info.owner != target_data->owner) {
            if (map_info.owner)
                /* don't warn for FF targets. */
                console.write(LOG_WARN,
                              format(_("Ship #%d has inconsistent XY information - correcting."), map_info.id));
            map_info.pos.x = target_data->x;
            map_info.pos.y = target_data->y;
            map_info.owner = target_data->owner;
        }

        /* create ship_data entry */
        copyTargetToShip(createShip(), *target_data);
    } else if (map_info.owner != 0) {
        /* nonvisual contact */
        map_info.kind = CurrentUnknown;

        /* create ship_data entry */
        createShip();
        ship_data->x     = map_info.pos.x;
        ship_data->y     = map_info.pos.y;
        ship_data->owner = map_info.owner;
    } else if (ship_data != 0) {
        /* history ship */
        map_info.kind = HistoryShip;
    } else {
        /* unknown */
        map_info.kind = NoShip;
    }
}

void
GShip::combinedCheck1(GUniverse& univ)
{
    if (isPlayed() && !hasShipData()) {
        /* we own it but can't see it? impossible! */
        delete target_data;
        target_data = 0;
        map_info.pos.x = map_info.pos.y = map_info.owner = map_info.scanned_mass = 0;
        console.write(LOG_WARN,
                      format(_("Ship #%d is ours but we don't see it -- deleted."), map_info.id));
        // FIXME: what if during combined check someone else sees that we play this ship?
        setIsPlayed(false);
        map_info.kind = ship_data ? HistoryShip : NoShip;
        return;
    }

    /* This ship looks valid. Compute orbit. */
    map_info.orbit = univ.getPlanetAt(getPos());
    if (map_info.orbit)
        univ.getPlanet(map_info.orbit).setShipInOrbit(*this);
}

/** Check whether we have full, playable data. */
bool
GShip::hasShipData() const
{
    return map_info.kind == CurrentShip;
}

/** Check whether we have any data about this ship. */
bool
GShip::hasAnyData() const
{
    return ship_data != 0;
}

/** Check whether this ship is reliably visible (to a player). A ship can
    be unreliably visible if it guessed. It can also be reliably visible to
    one player but not another one if they are not allied.

    \param for_player Player to ask question for: is this ship known
    to that player, and will host accept orders relating to it? If zero,
    check whether ship is seen reliably by anyone. */
bool
GShip::isReliablyVisible(int for_player) const
{
    GPlayerSet set = target_flags | src_flags;
    if (!for_player)
        return set.nonempty();
    else
        return set.contains(for_player);
}

/** Get kind of this ship. */
GShip::Kind
GShip::getKind() const
{
    return map_info.kind;
}

/** Get ship source flags. This is the set of players whose SHIP
    file contains a copy of this ship (usually a unit set). */
GPlayerSet
GShip::getSrcFlags() const
{
    return src_flags;
}

/** Get target source flags. This is the set of players whose TARGET
    file contains a copy of this ship, and who therefore reliably know
    about this ship's existance. Should be disjoint from getSrcFlags().

    FIXME: what about nonvisual contacts (SHIPXY)? */
GPlayerSet
GShip::getTargetFlags() const
{
    return target_flags;
}

/* Accessors */

/** Get ship mass. Returns zero if unknown. */
int
GShip::getMass() const
{
    if (map_info.kind == CurrentShip)
        return getShipMass(*ship_data);
    else if (map_info.kind == CurrentTarget || map_info.kind == CurrentUnknown)
        return map_info.scanned_mass;
    else
        return 0;
}

/** Get hull Id. Returns zero if unknown. */
int
GShip::getHullId() const
{
    // FIXME: convert to use mp16_t
    if (ship_data && ship_data->hull > 0)
        return ship_data->hull;
    else
        return 0;
}

/** Get hull object. */
GHull&
GShip::getHull() const
{
    int n = getHullId();
    ASSERT(n > 0);
    return ::getHull(n);
}

/** Get planet we're orbiting.
    \return planet id, 0 if none. */
int
GShip::getOrbit() const
{
    return map_info.orbit;
}

/** Get real owner of ship. */
int
GShip::getRealOwner() const
{
    if (map_info.rc_flag > 0)
        return map_info.rc_flag;
    else
        return getOwner();
}

/** Check whether this ship can do special function.
    \param basic_function A basic ship function, e.g. hf_Cloak. */
bool
GShip::canDoSpecial(int basic_function) const
{
    // FIXME: this is just a stub
    return getHull().canDoSpecial(basic_function, getRealOwner(), true,
                                  GExpLevelSet::allUpTo(NUM_EXPERIENCE_LEVELS));
}

/* Accessors: Course */

/** Get waypoint.
    \pre hasShipData() */
GPoint
GShip::getWaypoint() const
{
    ASSERT(map_info.kind == CurrentShip);
    return GPoint(map_info.pos.x + ship_data->waypoint_dx,
                  map_info.pos.y + ship_data->waypoint_dy);
}

/** Set waypoint.
    \pre hasShipData() */
void
GShip::setWaypoint(GPoint pt)
{
    if (ship_data != 0 && pt != getWaypoint()) {
        ship_data->waypoint_dx = pt.x - map_info.pos.x;
        ship_data->waypoint_dy = pt.y - map_info.pos.y;
        setIsDirty(ss_DirtyData);  // FIXME: DirtyXY?
    }
}

/** Get waypoint X displacement.
    \pre hasShipData() */
int
GShip::getWaypointDX() const
{
    ASSERT(map_info.kind == CurrentShip);
    return ship_data->waypoint_dx;
}

/** Get waypoint Y displacement.
    \pre hasShipData() */
int
GShip::getWaypointDY() const
{
    ASSERT(map_info.kind == CurrentShip);
    return ship_data->waypoint_dy;
}

/** Get ship's heading vector.
    \return value from [0,360), corresponding to VGAP's heading angles.
    Returns unknown value if ship doesn't move or heading not known */
mp16_t
GShip::getHeading() const
{
    if (map_info.kind == CurrentShip) {
        if (ship_data->waypoint_dx == 0 && ship_data->waypoint_dy == 0) {
            return mp16_t();
        } else {
            return mp16_t(int(getHeadingDeg(ship_data->waypoint_dx, ship_data->waypoint_dy)));
        }
    } else if (map_info.kind == CurrentTarget) {
        return mp16_t(target_data->heading);
    } else {
        return mp16_t();
    }
}

/** Get warp factor.
    \return warp factor or unknown */
mp16_t
GShip::getWarp() const
{
    if (map_info.kind == CurrentShip || map_info.kind == CurrentTarget)
        return mp16_t(ship_data->warp);
    else
        return mp16_t();
}

/** Set warp factor. */
void
GShip::setWarp(int warp)
{
    // FIXME: doesn't create history ships - should it?
    if (ship_data != 0 && warp != getWarp()) {
        ship_data->warp = warp;
        setIsDirty(ss_DirtyData);
    }
}

/** Check whether ship is hyperdriving.
    \return true iff ship has hyperdrive and tries to activate it */
bool
GShip::isHyperdriving() const
{
    return map_info.kind == CurrentShip
        && canDoSpecial(hf_Hyperdrive)
        && ship_data->warp > 0
        && std::memcmp("HYP", ship_data->fcode, 3) == 0;
}

/* Accessors: equipment */

/** Get engine type.
    \return engine type [1,NUM_ENGINES], or unknown */
mp16_t
GShip::getEngine() const
{
    return mp16_t(getDisplayedShip().engine);
}

/** Set engine type (for history).
    \param engine engine type [1,NUM_ENGINES], or unknown */
void
GShip::setEngine(mp16_t engine)
{
    createShip().engine = engine.getRawValue();
    setIsDirty(ss_DirtyData);
}

/** Get beam type.
    \return beam type [1,NUM_BEAMS], or unknown */
mp16_t
GShip::getBeamType() const
{
    return mp16_t(getDisplayedShip().beam_type);
}

/** Set beam type (for history).
    \param type beam type [1,NUM_BEAMS], or unknown */
void
GShip::setBeamType(mp16_t type)
{
    createShip().beam_type = type.getRawValue();
    setIsDirty(ss_DirtyData);
}

/** Get beam count.
    \return beam count, or unknown */
mp16_t
GShip::getBeamCount() const
{
    return mp16_t(getDisplayedShip().beam_count);
}

/** Set beam count (for history).
    \param count beam count, or unknown */
void
GShip::setBeamCount(mp16_t count)
{
    createShip().beam_count = count.getRawValue();
    setIsDirty(ss_DirtyData);
}

/** Get bay count.
    \return bay count, or unknown */
mp16_t
GShip::getBayCount() const
{
    return mp16_t(getDisplayedShip().bay_count);
}

/** Set bay count (for history).
    \param count bay count, or unknown */
void
GShip::setBayCount(mp16_t count)
{
    createShip().bay_count = count.getRawValue();
    setIsDirty(ss_DirtyData);
}

/** Get torpedo type.
    \return torpedo type [1,NUM_TORPS], or unknown */
mp16_t
GShip::getTorpType() const
{
    return mp16_t(getDisplayedShip().torp_type);
}

/** Set torpedo type (for history).
    \param type beam type [1,NUM_TORPS], or unknown */
void
GShip::setTorpType(mp16_t type)
{
    createShip().torp_type = type.getRawValue();
    setIsDirty(ss_DirtyData);
}

/** Get torpedo launcher count.
    \return torpedo launcher count, or unknown */
mp16_t
GShip::getTorpLauncherCount() const
{
    return mp16_t(getDisplayedShip().torp_launcher_count);
}

/** Set torpedo launcher count (for history).
    \param count torpedo launcher count, or unknown */
void
GShip::setTorpLauncherCount(mp16_t count)
{
    createShip().torp_launcher_count = count.getRawValue();
    setIsDirty(ss_DirtyData);
}

/* Accessors: Mission */

/** Set ship name.
    \param str New name */
void
GShip::setName(const string_t& str)
{
    storeBasicString(createShip().name, str);
    setIsDirty(ss_DirtyHead);
}

/** Get ship mission.
    \return mission number, or unknown */
mp16_t
GShip::getMission() const
{
    if (hasShipData())
        return mp16_t(getDisplayedShip().mission);
    else
        return mp16_t();
}

/** Set ship mission.
    \param m Mission number
    \param i Intercept number
    \param t Tow number

    FIXME: should this take a TMissionData object? */
void
GShip::setMission(int m, int i, int t)
{
    // FIXME: doesn't create history ships - should it? Should this work at all on history ships?
    if (ship_data == 0)
        return;

    if (m != ship_data->mission || i != ship_data->mission_intercept || t != ship_data->mission_tow) {
        ship_data->mission = m;
        ship_data->mission_intercept = i;
        ship_data->mission_tow = t;
        setIsDirty(ss_DirtyData);

        /* FIXME: I do not exactly like this, because it's slow, and
           it references getDisplayedTurn(). Can we do better? */
        if (const GMission* msn = GMissionList::getInstance().getMissionByNumber(m, GPlayerSet(raceId(getRealOwner())))) {
            if (msn->getFlags() & GMission::mf_Waypoint) {
                if (i > 0 && i <= NUM_SHIPS && getDisplayedTurn().getShip(i).isExistant()) {
                    setWaypoint(getDisplayedTurn().getShip(i).getPos());
                }
            }
        }
    }
}

/** Get tow number. */
mp16_t
GShip::getTowId() const
{
    if (hasShipData())
        return mp16_t(getDisplayedShip().mission_tow);
    else
        return mp16_t();
}

/** Get intercept number. */
mp16_t
GShip::getIntId() const
{
    if (hasShipData())
        return mp16_t(getDisplayedShip().mission_intercept);
    else
        return mp16_t();
}

/** Get primary enemy. */
mp16_t
GShip::getPE() const
{
    return mp16_t(getDisplayedShip().primary_enemy);
}

/** Set primary enemy.
    \param pe New primary enemy, [0,NUM_PLAYERS] */
void
GShip::setPE(int pe)
{
    // FIXME: doesn't create history ships - should it?
    if (ship_data != 0 && ship_data->primary_enemy != pe) {
        ship_data->primary_enemy = pe;
        setIsDirty(ss_DirtyData);
    }
}

/** Get ship damage.
    \return damage level, or unknown */
mp16_t
GShip::getDamage() const
{
    return mp16_t(getDisplayedShip().damage);
}

/** Set ship damage.
    \param damage damage level, or unknown */
void
GShip::setDamage(mp16_t damage)
{
    createShip().damage = damage.getRawValue();
    setIsDirty(ss_DirtyData);
}

/** Get ship crew.
    \return number of crewmen, or unknown */
mp16_t
GShip::getCrew() const
{
    return mp16_t(getDisplayedShip().crew);
}

/** Set ship crew.
    \param crew number of crewmen, or unknown */
void
GShip::setCrew(mp16_t crew)
{
    createShip().crew = crew.getRawValue();
    setIsDirty(ss_DirtyData);
}

/** Get friendly code. */
string_t
GShip::getFCode() const
{
    return getBasicString(getDisplayedShip().fcode);
}

/** Set friendly code. */
void
GShip::setFCode(const string_t& str)
{
    storeBasicString(createShip().fcode, str);
    setIsDirty(ss_DirtyData);
}

/* Accessors: cargo */

/* FIXME: how to handle cargo of history ships? We return int for the
   time being, as that's what planets do. */

int
GShip::getAmmo() const
{
    // FIXME: handle overload etc?
    GCargoType type = getBayCount() > 0 ? el_Fighters : GCargoType(el_Torps + getTorpType() - 1);
    return ship_data->ammo - cargo.getReservedCargo(type);
}

int
GShip::getColonists() const
{
    return getCargo(el_Colonists);
}

int
GShip::getSupplies() const
{
    return getCargo(el_Supplies);
}

int
GShip::getMoney() const
{
    return getCargo(el_Money);
}

/** Get cargo amount sans reservations. */
int
GShip::getCargo(GCargoType type) const
{
    return getCargoRaw(type) - cargo.getReservedCargo(type);
}

/** Get cargo which is physically aboard the ship. */
int
GShip::getCargoRaw(GCargoType type) const
{
    if (!ship_data)
        return 0;

    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return ship_data->ore_cargo[type];
     case el_Fighters:
        if (ship_data->bay_count > 0)
            return ship_data->ammo;
        return 0;
     case el_Colonists:
        return ship_data->colonists;
     case el_Supplies:
        return ship_data->supplies;
     case el_Money:
        return ship_data->money;
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        if (ship_data->torp_type > 0 && getTorpType() == (type - el_Torps + 1))
            return ship_data->ammo;
        return 0;
    }
}

/** Change amount that is on ship. */
void
GShip::changeCargoRaw(GCargoType type, int amount)
{
    if (!amount)
        return;

    // FIXME: we need a way to directly set a ship's cargo, for
    // foreign ships. This is needed for various scans:
    // - web drain complete (=set N to 0)
    // - combat result (=set ammo)
    ASSERT(hasShipData());
    setIsDirty(ss_DirtyData);
    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        ship_data->ore_cargo[type] += amount;
        break;
     case el_Fighters:
        ship_data->ammo += amount;
        break;
     case el_Colonists:
        ship_data->colonists += amount;
        break;
     case el_Supplies:
        ship_data->supplies += amount;
        break;
     case el_Money:
        ship_data->money += amount;
        break;
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        ship_data->ammo += amount;
        break;
    }
}

int
GShip::getFreeCargo() const
{
    if (map_info.kind == CurrentShip)
        return getHull().getCargo()
            - getCargo(el_Tritanium)
            - getCargo(el_Duranium)
            - getCargo(el_Molybdenum)
            - getAmmo()
            - getColonists()
            - getSupplies();
    else
        return 0;
}

/* Accessors: Transporters */

/** Check whether a transporter is active. An active transporter has a
    full transfer order. */
bool
GShip::isTransporterActive(Transporter which_one) const throw()
{
    if (hasShipData()) {
        const TShipTransfer& t = getTransporter(which_one);
        return (t.ore[0] | t.ore[1] | t.ore[2] | t.ore[3] | t.colonists | t.supplies) != 0;
    } else {
        return false;
    }
}

/** Check whether a transporter is cancellable. A transporter is
    cancellable if its content can be moved back into main cargo room
    without overloading the ship. */
bool
GShip::isTransporterCancellable(Transporter which_one) const throw()
{
    // FIXME: this does not guarantee correctness in all cases: Take a
    // ship with 10 cargo room left, and 10 kt in the transporter. A
    // pending transaction adds 10 to the ship (i.e. it got the
    // guarantee that it'll be able to add this). When we now do
    // cancelTranporter() and commit that transaction, the ship is
    // overloaded.
    if (hasShipData()) {
        const TShipTransfer& t = getTransporter(which_one);
        long sum = (long) t.ore[0] + t.ore[1] + t.ore[2] + t.ore[3] + t.supplies + t.colonists;
        return sum == 0 || sum <= getFreeCargo();
    } else {
        return false;
    }
}

/** Cancel a transporter. Moves its contents back to main cargo room.
    \pre isTransporterCancellable(which_one) */
void
GShip::cancelTranporter(Transporter which_one) throw()
{
    ASSERT(isTransporterCancellable(which_one));
    TShipTransfer& t = getTransporter(which_one);
    ship_data->ore_cargo[0] += t.ore[0];
    ship_data->ore_cargo[1] += t.ore[1];
    ship_data->ore_cargo[2] += t.ore[2];
    ship_data->ore_cargo[3] += t.ore[3];
    ship_data->colonists += t.colonists;
    ship_data->supplies += t.supplies;
    t.ore[0] = t.ore[1] = t.ore[2] = t.ore[3] = t.supplies = t.colonists = t.id = 0;

    setIsDirty(ss_DirtyData);
}

/** Check whether transporter is in use. A transporter is in use if
    the user is currently preparing a transport order. This means we
    must deny all other attempts of starting to prepare such an
    order. */
bool
GShip::isTransporterInUse(Transporter which_one) const throw()
{
    return transporter_in_use[which_one];
}

/** Set transporter in-use mark.
    \see isTransporterInUse */
void
GShip::setTransporterInUse(Transporter which_one, bool in_use) throw()
{
    transporter_in_use[which_one] = in_use;
}

/** Check whether ship is allowed to change waypoint. */
bool
GShip::getWaypointPermission() const
{
    // FIXME: real fleet implementation here
    return true;
}

/** Check whether ship wants to be cloned. This does not whether
    cloning actually is permitted (registeredness, hull restrictions,
    player restrictions). */
bool
GShip::isCloningAt(const GPlanet& planet) const
{
    return ship_data
        && planet.getPos() == getPos()
        && (ship_data->fcode[0] == 'c' && ship_data->fcode[1] == 'l' && ship_data->fcode[2] == 'n');
}

void
GShip::leaveFleet()
{
    // FIXME: real implementation here
}

/******************************* GShipType *******************************/

GShipType::GShipType(GUniverse& univ)
    : GObjectType(NUM_SHIPS, univ)
{ }

bool
GShipType::isExistant(int index)
{
    if (index > 0 && index <= NUM_SHIPS) {
        GShip& s = getUniverse().getShip(index);
        if (s.isExistant() && s.isPlayed())
            return true;
    }
    return false;
}

GShip&
GShipType::getObject(int index)
{
    return getUniverse().getShip(index);
}

string_t
GShipType::getClassName()
{
    return string_t(_("Starship"));
}
